/* -*-c-*-
 *
 * $Id$
 *
 * Admin interface for configuration
 *
 * (c) 2001 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of Trivial IP Encryption (TrIPE).
 *
 * TrIPE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * TrIPE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with TrIPE; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*----- Header files ------------------------------------------------------*/

#include "tripe.h"

/*----- Global variables --------------------------------------------------*/

#ifndef NTRACE

const trace_opt tr_opts[] = {
  { 't',	T_TUNNEL,	"tunnel events" },
  { 'r',	T_PEER,		"peer events" },
  { 'a',	T_ADMIN,	"admin interface" },
  { 'p',	T_PACKET,	"packet contents" },
  { 'c',	T_CRYPTO,	"crypto details" },
  { 's',	T_KEYSET,	"symmetric keyset management" },
  { 'x',	T_KEYEXCH,	"key exchange" },
  { 'm',	T_KEYMGMT,	"key management" },
  { 'A',	T_ALL,		"all of the above" },
  { 0,		0,		0 }
};

unsigned tr_flags = 0;
#endif

static const trace_opt w_opts[] = {
  { 't',	AF_TRACE,	"trace messages" },
  { 'n',	AF_NOTE,	"asynchronous notifications" },
  { 'w',	AF_WARN,	"warnings" },
  { 'A',	AF_ALLMSGS,	"all of the above" },
  { 0,		0,		0 }
};

/*----- Static variables --------------------------------------------------*/

static admin *admins;
static sel_file sock;
static const char *sockname;
static unsigned flags = 0;
static admin *a_stdin = 0;
static sig s_term, s_int, s_hup;

#define F_DAEMON 1u
#define F_INIT 2u

#define T_RESOLVE SEC(30)

static void a_destroy(admin */*a*/);
static void a_lock(admin */*a*/);
static void a_unlock(admin */*a*/);

/*----- Output functions --------------------------------------------------*/

/* --- @mystrieq@ --- *
 *
 * Arguments:	@const char *x, *y@ = two strings
 *
 * Returns:	True if @x@ and @y are equal, up to case.
 */

static int mystrieq(const char *x, const char *y)
{
  for (;;) {
    if (!*x && !*y) return (1);
    if (tolower((unsigned char)*x) != tolower((unsigned char)*y))
      return (0);
    x++; y++;
  }
}

/* --- @trywrite@ --- *
 *
 * Arguments:	@admin *a@ = pointer to an admin block
 *		@const char *p@ = pointer to buffer to write
 *		@size_t sz@ = size of data to write
 *
 * Returns:	The number of bytes written, or less than zero on error.
 *
 * Use:		Attempts to write data to a client.
 */

static ssize_t trywrite(admin *a, const char *p, size_t sz)
{
  ssize_t n, done = 0;

again:
  if (!sz)
    return (done);
  n = write(a->w.fd, p, sz);
  if (n > 0) {
    done += n;
    p += n;
    sz -= n;
    goto again;
  }
  if (n < 0) {
    if (errno == EINTR)
      goto again;
    if (errno != EAGAIN && errno != EWOULDBLOCK) {
      a_destroy(a);
      a_warn("ADMIN client-read-error -- %s", strerror(errno));
      return (-1);
    }
  }
  return (done);
}

/* --- @dosend@ --- *
 *
 * Arguemnts:	@admin *a@ = pointer to an admin block
 *		@const char *p@ = pointer to buffer to write
 *		@size_t sz@ = size of data to write
 *
 * Returns:	---
 *
 * Use:		Sends data to an admin client.
 */

static void dosend(admin *a, const char *p, size_t sz)
{
  ssize_t n;
  obuf *o;

  if (a->f & AF_DEAD)
    return;

  /* --- Try to send the data immediately --- */

  if (!a->o_head) {
    if ((n = trywrite(a, p, sz)) < 0)
      return;
    p += n;
    sz -= n;
    if (!sz)
      return;
  }
	
  /* --- Fill buffers with the data until it's all gone --- */

  o = a->o_tail;
  if (!o)
    sel_addfile(&a->w);
  else if (o->p_in < o->buf + OBUFSZ)
    goto noalloc;

  do {
    o = xmalloc(sizeof(obuf));
    o->next = 0;
    o->p_in = o->p_out = o->buf;
    if (a->o_tail)
      a->o_tail->next = o;
    else
      a->o_head = o;
    a->o_tail = o;

  noalloc:
    n = o->buf + OBUFSZ - o->p_in;
    if (n > sz)
      n = sz;
    memcpy(o->p_in, p, n);
    o->p_in += n;
    p += n;
    sz -= n;
  } while (sz);
}

/* --- @a_flush@ --- *
 *
 * Arguments:	@int fd@ = file descriptor
 *		@unsigned mode@ = what's happening
 *		@void *v@ = pointer to my admin block
 *
 * Returns:	---
 *
 * Use:		Flushes buffers when a client is ready to read again.
 */

static void a_flush(int fd, unsigned mode, void *v)
{
  admin *a = v;
  obuf *o, *oo;
  ssize_t n;

  o = a->o_head;
  while (o) {
    if ((n = trywrite(a, o->p_out, o->p_in - o->p_out)) < 0)
      return;
    o->p_out += n;
    if (o->p_in < o->p_out)
      break;
    oo = o;
    o = o->next;
    xfree(oo);
  }
  a->o_head = o;
  if (!o) {
    a->o_tail = 0;
    sel_rmfile(&a->w);
  }
}

/*----- Utility functions -------------------------------------------------*/

/* --- @a_write@, @a_vwrite@ --- *
 *
 * Arguments:	@admin *a@ = admin connection to write to
 *		@const char *tag@ = tag prefix string, or null
 *		@const char *fmt@ = pointer to format string
 *		@va_list ap@ = arguments in list
 *		@...@ = other arguments
 *
 * Returns:	---
 *
 * Use:		Sends a message to an admin connection.
 */

static void a_vwrite(admin *a, const char *tag, const char *fmt, va_list ap)
{
  dstr d = DSTR_INIT;
  if (tag) {
    dstr_puts(&d, tag);
    if (fmt)
      dstr_putc(&d, ' ');
  }
  if (fmt)
    dstr_vputf(&d, fmt, &ap);
  dstr_putc(&d, '\n');
  dosend(a, d.buf, d.len);
  dstr_destroy(&d);
}

static void a_write(admin *a, const char *tag, const char *fmt, ...)
{
  va_list ap;
  va_start(ap, fmt);
  a_vwrite(a, tag, fmt, ap);
  va_end(ap);
}

/* --- @a_ok@, @a_info@, @a_fail@ --- *
 *
 * Arguments:	@admin *a@ = connection
 *		@const char *fmt@ = format string
 *		@...@ = other arguments
 *
 * Returns:	---
 *
 * Use:		Convenience functions for @a_write@.
 */

static void a_ok(admin *a) { a_write(a, "OK", 0); }

static void a_info(admin *a, const char *fmt, ...)
{
  va_list ap;
  va_start(ap, fmt);
  a_vwrite(a, "INFO", fmt, ap);
  va_end(ap);
}

static void a_fail(admin *a, const char *fmt, ...)
{
  va_list ap;
  va_start(ap, fmt);
  a_vwrite(a, "FAIL", fmt, ap);
  va_end(ap);
}

/* --- @a_alert@, @a_valert@, @a_rawalert@ --- *
 *
 * Arguments:	@unsigned f_and, f_eq@ = filter for connections
 *		@const char *tag@ = tag prefix string
 *		@const char *fmt@ = pointer to format string
 @		@const char *p@ = pointer to raw string
 *		@size_t sz@ = size of raw string
 *		@va_list ap@ = arguments in list
 *		@...@ = other arguments
 *
 * Returns:	---
 *
 * Use:		Write a message to all admin connections matched by the given
 *		filter.
 */

static void a_rawalert(unsigned f_and, unsigned f_eq, const char *tag,
		       const char *p, size_t sz)
{
  admin *a, *aa;
  dstr d = DSTR_INIT;
  
  if (!(flags & F_INIT))
    return;
  if (tag) {
    dstr_puts(&d, tag);
    if (p)
      dstr_putc(&d, ' ');
  }
  if (p)
    dstr_putm(&d, p, sz);
  dstr_putc(&d, '\n');
  p = d.buf;
  sz = d.len;
  for (a = admins; a; a = aa) {
    aa = a->next;
    if ((a->f & f_and) == f_eq)
      dosend(a, d.buf, d.len);
  }
  dstr_destroy(&d);
}

static void a_valert(unsigned f_and, unsigned f_eq, const char *tag,
		     const char *fmt, va_list ap)
{
  dstr d = DSTR_INIT;

  if (!(flags & F_INIT))
    return;
  if (fmt)
    dstr_vputf(&d, fmt, &ap);
  a_rawalert(f_and, f_eq, tag, fmt ? d.buf : 0, fmt ? d.len : 0);
  dstr_destroy(&d);
}

#if 0 /*unused*/
static void a_alert(unsigned f_and, unsigned f_eq, const char *tag,
		    const char *fmt, ...)
{
  va_list ap;
  va_start(ap, fmt);
  a_valert(f_and, f_eq, tag, fmt, ap);
  va_end(ap);
}
#endif

/* --- @a_warn@ --- *
 *
 * Arguments:	@const char *fmt@ = pointer to format string
 *		@...@ = other arguments
 *
 * Returns:	---
 *
 * Use:		Informs all admin connections of a warning.
 */

void a_warn(const char *fmt, ...)
{
  va_list ap;

  va_start(ap, fmt);
  if (flags & F_INIT)
    a_valert(0, 0, "WARN", fmt, ap);
  else {
    fprintf(stderr, "%s: ", QUIS);
    vfprintf(stderr, fmt, ap);
    fputc('\n', stderr);
  }
  va_end(ap);
}

/* --- @a_trace@ --- *
 *
 * Arguments:	@const char *p@ = pointer to a buffer
 *		@size_t sz@ = size of the buffer
 *		@void *v@ = uninteresting pointer
 *
 * Returns:	---
 *
 * Use:		Custom trace output handler.  Sends trace messages to
 *		interested admin connections.
 */

#ifndef NTRACE
static void a_trace(const char *p, size_t sz, void *v)
{
  a_rawalert(AF_TRACE, AF_TRACE, "TRACE", p, sz);
}
#endif

/* --- @a_notify@ --- *
 *
 * Arguments:	@const char *fmt@ = pointer to format string
 *		@...@ = other arguments
 *
 * Returns:	---
 *
 * Use:		Sends a notification to interested admin connections.
 */

void a_notify(const char *fmt, ...)
{
  va_list ap;

  va_start(ap, fmt);
  a_valert(AF_NOTE, AF_NOTE, "NOTE", fmt, ap);
  va_end(ap);
}

/* --- @a_quit@ --- *
 *
 * Arguments:	---
 *
 * Returns:	---
 *
 * Use:		Shuts things down nicely.
 */

void a_quit(void)
{
  close(sock.fd);
  unlink(sockname);
  exit(0);
}

/* --- @a_sigdie@ --- *
 *
 * Arguments:	@int sig@ = signal number
 *		@void *v@ = an uninteresting argument
 *
 * Returns:	---
 *
 * Use		Shuts down on receipt of a fatal signal.
 */

static void a_sigdie(int sig, void *v)
{
  char *p;
  char buf[20];

  switch (sig) {
    case SIGTERM:	p = "SIGTERM"; break;
    case SIGINT:	p = "SIGINT"; break;
    default:
      sprintf(buf, "%i", sig);
      p = buf;
      break;
  }
  a_warn("SERVER quit signal %s", p);
  a_quit();
}

/* --- @a_sighup@ --- *
 *
 * Arguments:	@int sig@ = signal number
 *		@void *v@ = an uninteresting argument
 *
 * Returns:	---
 *
 * Use		Logs a message about SIGHUP not being useful.
 */

static void a_sighup(int sig, void *v)
{
  a_warn("SERVER ignore signal SIGHUP");
}

/*----- Adding peers ------------------------------------------------------*/
 
/* --- @a_resolve@ --- *
 *
 * Arguments:	@struct hostent *h@ = pointer to resolved hostname
 *		@void *v@ = pointer to admin block
 *
 * Returns:	---
 *
 * Use:		Handles a completed name resolution.
 */

static void a_resolve(struct hostent *h, void *v)
{
  admin *a = v;

  a_lock(a);
  T( trace(T_ADMIN, "admin: %u resolved", a->seq); )
  TIMER;
  sel_rmtimer(&a->t);
  if (!h)
    a_fail(a, "resolve-error %s", a->paddr);
  else if (p_find(a->pname))
    a_fail(a, "peer-exists %s", a->pname);
  else {
    memcpy(&a->peer.sin.sin_addr, h->h_addr, sizeof(struct in_addr));
    if (!p_create(a->pname, &a->peer.sa, a->sasz))
      a_fail(a, "peer-create-fail %s" a->pname);
    else
      a_ok(a);
  }
  xfree(a->pname);
  xfree(a->paddr);
  a->pname = 0;
  selbuf_enable(&a->b);
  a_unlock(a);
}

/* --- @a_timer@ --- *
 *
 * Arguments:	@struct timeval *tv@ = timer
 *		@void *v@ = pointer to admin block
 *
 * Returns:	---
 *
 * Use:		Times out a resolver.
 */

static void a_timer(struct timeval *tv, void *v)
{
  admin *a = v;

  a_lock(a);
  T( trace(T_ADMIN, "admin: %u resolver timeout", a->seq); )
  bres_abort(&a->r);
  a_fail(a, "resolver-timeout %s\n", a->paddr);
  xfree(a->pname);
  xfree(a->paddr);
  a->pname = 0;
  selbuf_enable(&a->b);
  a_unlock(a);
}

/* --- @acmd_add@ --- *
 *
 * Arguments:	@admin *a@ = connection which requested the addition
 *		@unsigned ac@ = argument count
 *		@char *av[]@ = pointer to the argument list
 *
 * Returns:	---
 *
 * Use:		Adds a new peer.
 */

static void acmd_add(admin *a, unsigned ac, char *av[])
{
  unsigned long pt;
  struct timeval tv;
  unsigned i;
  char *p;

  /* --- Make sure someone's not got there already --- */

  if (p_find(av[0])) {
    a_fail(a, "peer-exists %s", av[0]);
    return;
  }

  /* --- Fill in the easy bits of address --- */

  BURN(a->peer);
  i = 1;
  if (mystrieq(av[i], "inet")) i++;
  if (ac - i != 2) {
    a_fail(a, "bad-syntax -- add PEER [inet] ADDRESS PORT");
    return;
  }
  a->peer.sin.sin_family = AF_INET;
  a->sasz = sizeof(a->peer.sin);
  pt = strtoul(av[i + 1], &p, 0);
  if (*p) {
    struct servent *s = getservbyname(av[i + 1], "udp");
    if (!s) {
      a_fail(a, "unknown-service %s", av[i + 1]);
      return;
    }
    pt = ntohs(s->s_port);
  }
  if (pt == 0 || pt >= 65536) {
    a_fail(a, "invalid-port %lu", pt);
    return;
  }
  a->peer.sin.sin_port = htons(pt);

  /* --- If the name is numeric, do it the easy way --- */
  
  if (inet_aton(av[i], &a->peer.sin.sin_addr)) {
    if (!p_create(av[0], &a->peer.sa, a->sasz))
      a_fail(a, "peer-create-fail %s", a->pname);
    else
      a_ok(a);
    return;
  }

  /* --- Store everything for later and crank up the resolver --- *
   *
   * We disable the line buffer until the resolver completes (or times out).
   * This prevents other commands on the same connection (though the rest of
   * the system continues regardless), but makes life simpler for the client.
   */

  a->pname = xstrdup(av[0]);
  a->paddr = xstrdup(av[i]);
  selbuf_disable(&a->b);
  gettimeofday(&tv, 0);
  tv.tv_sec += T_RESOLVE;
  sel_addtimer(&sel, &a->t, &tv, a_timer, a);
  bres_byname(&a->r, a->paddr, a_resolve, a);
  T( trace(T_ADMIN, "admin: %u resolving hostname `%s'",
	   a->seq, a->paddr); )
}

/*----- Administration commands -------------------------------------------*/

/* --- Miscellaneous commands --- */

/* --- @traceish@ --- *
 *
 * Arguments:	@admin *a@ = connection to complain on
 *		@unsigned ac@ = number of arguments
 *		@char *av[]@ = vector of arguments
 *		@const char *what@ = what we're messing with
 *		@const trace_opt *tt@ = options table
 *		@unsigned *ff@ = where the flags are
 *
 * Returns:	Nonzero if anything changed.
 *
 * Use:		Guts of trace-ish commands like `trace' and `watch'.
 */

static int traceish(admin *a, unsigned ac, char *av[],
		    const char *what, const trace_opt *tt, unsigned *ff)
{
  int ch = 0;

  if (!ac || strcmp(av[0], "?") == 0) {
    const trace_opt *t;
    a_info(a, "Current %s status:", what);
    for (t = tt; t->ch; t++) {
      a_info(a, "%c %c  %s",
	     t->ch, (*ff & t->f) == t->f ? '*' : ' ', t->help);
    }
  } else {
    unsigned sense = 1;
    unsigned f = *ff;
    const trace_opt *t;
    char *p = av[0];

    while (*p) {
      switch (*p) {
	case '+': sense = 1; break;
	case '-': sense = 0; break;
	default:
	  for (t = tt; t->ch; t++) {
	    if (t->ch == *p) {
	      if (sense) f |= t->f;
	      else f &= ~t->f;
	      goto tropt_ok;
	    }
	  }
	  a_fail(a, "bad-%s-option %c", what, *p);
	  return (0);
        tropt_ok:;
	  break;
      }
      p++;
    }
    *ff = f;
    ch = 1;
  }
  a_ok(a);
  return (ch);
}

#ifndef NTRACE

static void acmd_trace(admin *a, unsigned ac, char *av[])
{
  if (traceish(a, ac, av, "trace", tr_opts, &tr_flags))
    trace_level(tr_flags);
}

#endif

static void acmd_watch(admin *a, unsigned ac, char *av[])
{
  traceish(a, ac, av, "watch", w_opts, &a->f);
}

static void quotify(dstr *d, const char *p)
{
  if (d->len)
    dstr_putc(d, ' ');
  if (*p && !p[strcspn(p, "\"' \t\n\v")])
    dstr_puts(d, p);
  else {
    dstr_putc(d, '\"');
    while (*p) {
      if (*p == '\\' || *p == '\"')
	dstr_putc(d, '\\');
      dstr_putc(d, *p++);
    }
    dstr_putc(d, '\"');
  }
}

static void alertcmd(admin *a, unsigned f_and, unsigned f_eq,
		     const char *tag, unsigned ac, char *av[])
{
  dstr d = DSTR_INIT;
  unsigned i;

  dstr_puts(&d, "USER");
  for (i = 0; i < ac; i++)
    quotify(&d, av[i]);
  dstr_putz(&d);
  a_rawalert(f_and, f_eq, tag, d.buf, d.len);
  dstr_destroy(&d);
  a_ok(a);
}

static void acmd_notify(admin *a, unsigned ac, char *av[])
  { alertcmd(a, AF_NOTE, AF_NOTE, "NOTE", ac, av); }
static void acmd_warn(admin *a, unsigned ac, char *av[])
  { alertcmd(a, AF_WARN, AF_WARN, "WARN", ac, av); }

static void acmd_port(admin *a, unsigned ac, char *av[])
{
  a_info(a, "%u", p_port());
  a_ok(a);
}

static void acmd_daemon(admin *a, unsigned ac, char *av[])
{
  if (flags & F_DAEMON)
    a_fail(a, "already-daemon");
  else {
    a_notify("DAEMON");
    if (a_stdin)
      a_destroy(a_stdin);
    if (u_daemon())
      a_fail(a, "daemon-error -- %s", strerror(errno));
    else {
      flags |= F_DAEMON;
      a_ok(a);
    }
  }
}

static void acmd_list(admin *a, unsigned ac, char *av[])
{
  peer *p;
  for (p = p_first(); p; p = p_next(p))
    a_info(a, "%s", p_name(p));
  a_ok(a);
}

static void acmd_ifname(admin *a, unsigned ac, char *av[])
{
  peer *p;

  if ((p = p_find(av[0])) == 0)
    a_fail(a, "unknown-peer %s", av[0]);
  else {
    a_info(a, "%s", p_ifname(p));
    a_ok(a);
  }
}

static void acmd_addr(admin *a, unsigned ac, char *av[])
{
  peer *p;
  const addr *ad;

  if ((p = p_find(av[0])) == 0)
    a_fail(a, "unknown-peer %s", av[0]);
  else {
    ad = p_addr(p);
    assert(ad->sa.sa_family == AF_INET);
    a_info(a, "INET %s %u",
	    inet_ntoa(ad->sin.sin_addr),
	    (unsigned)ntohs(ad->sin.sin_port));
    a_ok(a);
  }
}

static void acmd_stats(admin *a, unsigned ac, char *av[])
{
  peer *p;
  stats *st;

  if ((p = p_find(av[0])) == 0)
    a_fail(a, "unknown-peer %s", av[0]);
  else {
    st = p_stats(p);
    a_info(a, "start-time=%s", timestr(st->t_start));
    a_info(a, "last-packet-time=%s", timestr(st->t_last));
    a_info(a, "last-keyexch-time=%s", timestr(st->t_kx));
    a_info(a, "packets-in=%lu bytes-in=%lu", st->n_in, st->sz_in);
    a_info(a, "packets-out=%lu bytes-out=%lu",
	    st->n_out, st->sz_out);
    a_info(a, "keyexch-packets-in=%lu keyexch-bytes-in=%lu",
	    st->n_kxin, st->sz_kxin);
    a_info(a, "keyexch-packets-out=%lu keyexch-bytes-out=%lu",
	    st->n_kxout, st->sz_kxout);
    a_info(a, "ip-packets-in=%lu ip-bytes-in=%lu",
	    st->n_ipin, st->sz_ipin);
    a_info(a, "ip-packets-out=%lu ip-bytes-out=%lu",
	    st->n_ipout, st->sz_ipout);
    a_info(a, "rejected-packets=%lu", st->n_reject);
    a_ok(a);
  }
}

static void acmd_kill(admin *a, unsigned ac, char *av[])
{
  peer *p;
  if ((p = p_find(av[0])) == 0)
    a_fail(a, "unknown-peer %s", av[0]);
  else {
    p_destroy(p);
    a_ok(a);
  }
}

static void acmd_quit(admin *a, unsigned ac, char *av[])
{
  a_warn("SERVER quit admin-request");
  a_ok(a);
  a_quit();
}

static void acmd_version(admin *a, unsigned ac, char *av[])
{
  a_info(a, "%s %s", PACKAGE, VERSION);
  a_ok(a);
}

/* --- The command table and help --- */

typedef struct acmd {
  const char *name;
  const char *help;
  unsigned argmin, argmax;
  void (*func)(admin */*a*/, unsigned /*ac*/, char */*av*/[]);
} acmd;

static void acmd_help(admin */*a*/, unsigned /*ac*/, char */*av*/[]);

static const acmd acmdtab[] = {
  { "help",	"help",			0,	0,	acmd_help },
  { "version",	"version",		0,	0,	acmd_version },
#ifndef NTRACE
  { "trace",	"trace [OPTIONS]",	0,	1,	acmd_trace },
#endif
  { "watch",	"watch [OPTIONS]", 	0,	1,	acmd_watch },
  { "notify",	"notify MESSAGE ...",	1,	0xffff,	acmd_notify },
  { "warn",	"warn MESSAGE ...",	1,	0xffff,	acmd_warn },
  { "port",	"port",			0,	0,	acmd_port },
  { "daemon",	"daemon",		0,	0,	acmd_daemon },
  { "list",	"list",			0,	0,	acmd_list },
  { "ifname",	"ifname PEER",		1,	1,	acmd_ifname },
  { "addr",	"addr PEER",		1,	1,	acmd_addr },
  { "stats",	"stats PEER",		1,	1,	acmd_stats },
  { "kill",	"kill PEER",		1,	1,	acmd_kill },
  { "add",	"add PEER ADDR ...",	2,	0xffff,	acmd_add },
  { "quit",	"quit",			0,	0,	acmd_quit },
  { 0,		0,			0,	0,	0 }
};

static void acmd_help(admin *a, unsigned ac, char *av[])
{
  const acmd *c;
  for (c = acmdtab; c->name; c++)
    a_info(a, "%s", c->help);
  a_ok(a);
}

/*----- Connection handling -----------------------------------------------*/

/* --- @a_lock@ --- *
 *
 * Arguments:	@admin *a@ = pointer to an admin block
 *
 * Returns:	---
 *
 * Use:		Locks an admin block so that it won't be destroyed
 *		immediately.
 */

static void a_lock(admin *a) { assert(!(a->f & AF_LOCK)); a->f |= AF_LOCK; }

/* --- @a_unlock@ --- *
 *
 * Arguments:	@admin *a@ = pointer to an admin block
 *
 * Returns:	---
 *
 * Use:		Unlocks an admin block, allowing its destruction.  This is
 *		also the second half of @a_destroy@.
 */

static void a_unlock(admin *a)
{
  assert(a->f & AF_LOCK);
  if (!(a->f & AF_DEAD)) {
    a->f &= ~AF_LOCK;
    return;
  }

  T( trace(T_ADMIN, "admin: completing destruction of connection %u",
	   a->seq); )

  selbuf_destroy(&a->b);
  if (a->pname) {
    xfree(a->pname);
    xfree(a->paddr);
    bres_abort(&a->r);
    sel_rmtimer(&a->t);
  }
  if (a->b.reader.fd != a->w.fd)
    close(a->b.reader.fd);
  close(a->w.fd);

  if (a_stdin == a)
    a_stdin = 0;
  if (a->next)
    a->next->prev = a->prev;
  if (a->prev)
    a->prev->next = a->next;
  else
    admins = a->next;
  DESTROY(a);
}

/* --- @a_destroy@ --- *
 *
 * Arguments:	@admin *a@ = pointer to an admin block
 *
 * Returns:	---
 *
 * Use:		Destroys an admin block.  This requires a certain amount of
 *		care.
 */

static void a_destroy(admin *a)
{
  /* --- Don't multiply destroy admin blocks --- */

  if (a->f & AF_DEAD)
    return;

  /* --- Make sure nobody expects it to work --- */

  a->f |= AF_DEAD;
  T( trace(T_ADMIN, "admin: destroying connection %u", a->seq); )

  /* --- Free the output buffers --- */

  if (a->o_head) {
    obuf *o, *oo;
    sel_rmfile(&a->w);
    for (o = a->o_head; o; o = oo) {
      oo = o->next;
      xfree(o);
    }
    a->o_head = 0;
  }

  /* --- If the block is locked, that's all we can manage --- */

  if (a->f & AF_LOCK) {
    T( trace(T_ADMIN, "admin: deferring destruction..."); )
    return;
  }
  a->f |= AF_LOCK;
  a_unlock(a);
}

/* --- @a_line@ --- *
 *
 * Arguments:	@char *p@ = pointer to the line read
 *		@size_t len@ = length of the line
 *		@void *vp@ = pointer to my admin block
 *
 * Returns:	---
 *
 * Use:		Handles a line of input.
 */

static void a_line(char *p, size_t len, void *vp)
{
  admin *a = vp;
  const acmd *c;
  char *av[16];
  size_t ac;

  TIMER;
  if (a->f & AF_DEAD)
    return;
  if (!p) {
    a_destroy(a);
    return;
  }
  ac = str_qsplit(p, av, 16, 0, STRF_QUOTE);
  if (!ac)
    return;
  for (c = acmdtab; c->name; c++) {
    if (mystrieq(av[0], c->name)) {
      ac--;
      if (c->argmin > ac || ac > c->argmax)
	a_fail(a, "bad-syntax -- %s", c->help);
      else {
	a_lock(a);
	c->func(a, ac, av + 1);
	a_unlock(a);
      }
      return;
    }
  }
  a_fail(a, "unknown-command %s", av[0]);
}

/* --- @a_create@ --- *
 *
 * Arguments:	@int fd_in, fd_out@ = file descriptors to use
 *		@unsigned f@ = initial flags to set
 *
 * Returns:	---
 *
 * Use:		Creates a new admin connection.
 */

void a_create(int fd_in, int fd_out, unsigned f)
{
  admin *a = CREATE(admin);

  T( static unsigned seq = 0;
     a->seq = seq++; )
  T( trace(T_ADMIN, "admin: accepted connection %u", a->seq); )
  a->pname = 0;
  a->f = f;
  if (fd_in == STDIN_FILENO)
    a_stdin = a;
  fdflags(fd_in, O_NONBLOCK, O_NONBLOCK, FD_CLOEXEC, FD_CLOEXEC);
  if (fd_out != fd_in)
    fdflags(fd_out, O_NONBLOCK, O_NONBLOCK, FD_CLOEXEC, FD_CLOEXEC);
  selbuf_init(&a->b, &sel, fd_in, a_line, a);
  sel_initfile(&sel, &a->w, fd_out, SEL_WRITE, a_flush, a);
  a->o_head = 0;
  a->o_tail = 0;
  a->next = admins;
  a->prev = 0;
  if (admins)
    admins->prev = a;
  admins = a;
}

/* --- @a_accept@ --- *
 *
 * Arguments:	@int fd@ = file descriptor to accept
 *		@unsigned mode@ = what to do
 *		@void *v@ = uninteresting pointer
 *
 * Returns:	---
 *
 * Use:		Accepts a new admin connection.
 */

static void a_accept(int fd, unsigned mode, void *v)
{
  int nfd;
  struct sockaddr_un sun;
  size_t sz = sizeof(sun);

  if ((nfd = accept(fd, (struct sockaddr *)&sun, &sz)) < 0) {
    if (errno != EINTR && errno != EAGAIN && errno != EWOULDBLOCK &&
	errno != ECONNABORTED && errno != EPROTO)
      a_warn("ADMIN accept-error -- %s", strerror(errno));
    return;
  }
  a_create(nfd, nfd, 0);
}

/* --- @a_daemon@ --- *
 *
 * Arguments:	---
 *
 * Returns:	---
 *
 * Use:		Informs the admin module that it's a daemon.
 */

void a_daemon(void)
{
  flags |= F_DAEMON;
}

/* --- @a_init@ --- *
 *
 * Arguments:	@const char *name@ = socket name to create
 *
 * Returns:	---
 *
 * Use:		Creates the admin listening socket.
 */

void a_init(const char *name)
{
  int fd;
  int n = 5;
  struct sockaddr_un sun;
  struct sigaction sa;
  size_t sz;

  /* --- Set up the socket address --- */

  sz = strlen(name) + 1;
  if (sz > sizeof(sun.sun_path))
    die(EXIT_FAILURE, "socket name `%s' too long", name);
  BURN(sun);
  sun.sun_family = AF_UNIX;
  memcpy(sun.sun_path, name, sz);
  sz += offsetof(struct sockaddr_un, sun_path);

  /* --- Attempt to bind to the socket --- */

  umask(0077);
again:
  if ((fd = socket(PF_UNIX, SOCK_STREAM, 0)) < 0)
    die(EXIT_FAILURE, "couldn't create socket: %s", strerror(errno));
  if (bind(fd, (struct sockaddr *)&sun, sz) < 0) {
    struct stat st;
    int e = errno;
    if (errno != EADDRINUSE) {
      die(EXIT_FAILURE, "couldn't bind to address `%s': %s",
	  sun.sun_path, strerror(e));
    }
    if (!n)
      die(EXIT_FAILURE, "too many retries; giving up");
    n--;
    if (!connect(fd, (struct sockaddr *)&sun, sz)) {
      die(EXIT_FAILURE, "server already listening on admin socket `%s'",
	  sun.sun_path);
    }
    if (errno != ECONNREFUSED)
      die(EXIT_FAILURE, "couldn't bind to address: %s", strerror(e));
    if (stat(sun.sun_path, &st)) {
      die(EXIT_FAILURE, "couldn't stat `%s': %s",
	  sun.sun_path, strerror(errno));
    }
    if (!S_ISSOCK(st.st_mode))
      die(EXIT_FAILURE, "object `%s' isn't a socket", sun.sun_path);
    T( trace(T_ADMIN, "admin: stale socket found; removing it"); )
    unlink(sun.sun_path);
    close(fd);
    goto again;
  }
  chmod(sun.sun_path, 0600);
  fdflags(fd, O_NONBLOCK, O_NONBLOCK, FD_CLOEXEC, FD_CLOEXEC);
  if (listen(fd, 5))
    die(EXIT_FAILURE, "couldn't listen on socket: %s", strerror(errno));

  /* --- Listen to the socket --- */

  sel_initfile(&sel, &sock, fd, SEL_READ, a_accept, 0);
  sel_addfile(&sock);
  sockname = name;
  bres_init(&sel);
  T( trace_custom(a_trace, 0);
     trace(T_ADMIN, "admin: enabled custom tracing"); )
  flags |= F_INIT;

  /* --- Set up signal handlers --- */

  sig_add(&s_term, SIGTERM, a_sigdie, 0);
  sig_add(&s_hup, SIGHUP, a_sighup, 0);
  signal(SIGPIPE, SIG_IGN);
  sigaction(SIGINT, 0, &sa);
  if (sa.sa_handler != SIG_IGN)
    sig_add(&s_int, SIGINT, a_sigdie, 0);
}

/*----- That's all, folks -------------------------------------------------*/
