/* -*-c-*-
 *
 * $Id$
 *
 * Communication with the peer
 *
 * (c) 2001 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of Trivial IP Encryption (TrIPE).
 *
 * TrIPE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * TrIPE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with TrIPE; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*----- Header files ------------------------------------------------------*/

#include "tripe.h"

/*----- Static variables --------------------------------------------------*/

static peer *peers = 0;
static sel_file sock;

/*----- Main code ---------------------------------------------------------*/

/* --- @p_read@ --- *
 *
 * Arguments:	@int fd@ = file descriptor to read from
 *		@unsigned mode@ = what happened
 *		@void *v@ = an uninteresting pointer
 *
 * Returns:	---
 *
 * Use:		Reads a packet from somewhere.
 */

static void p_read(int fd, unsigned mode, void *v)
{
  peer *p;
  addr a;
  size_t sz;
  ssize_t n;
  int ch;
  buf b, bb;

  /* --- Read the data --- */

  TIMER;
  sz = sizeof(addr);
  n = recvfrom(fd, buf_i, sizeof(buf_i), 0, &a.sa, &sz);
  if (n < 0) {
    a_warn("PEER - socket-read-error -- %s", strerror(errno));
    return;
  }

  /* --- Find the appropriate peer --- */

  assert(a.sa.sa_family == AF_INET);
  for (p = peers; p; p = p->next) {
    if (p->peer.sin.sin_addr.s_addr == a.sin.sin_addr.s_addr &&
	p->peer.sin.sin_port == a.sin.sin_port)
      goto found;
  }
  a_warn("PEER - unexpected-source INET %s %u",
	 inet_ntoa(a.sin.sin_addr), (unsigned)ntohs(a.sin.sin_port));
  return;

found:
  T( trace(T_PEER, "peer: packet received from `%s'", p->name);
     trace_block(T_PACKET, "peer: packet contents", buf_i, n); )

  /* --- Pick the packet apart --- */

  p->st.t_last = time(0);
  p->st.n_in++;
  p->st.sz_in += n;
  buf_init(&b, buf_i, n);
  if ((ch = buf_getbyte(&b)) < 0) {
    a_warn("PEER %s bad-packet no-type", p->name);
    return;
  }
  switch (ch & MSG_CATMASK) {
    case MSG_PACKET:
      if (ch & MSG_TYPEMASK) {
	a_warn("PEER %s bad-packet unknown-type 0x%02x", p->name, ch);
	p->st.n_reject++;
	return;
      }
      buf_init(&bb, buf_o, sizeof(buf_o));
      if (ksl_decrypt(&p->ks, MSG_PACKET, &b, &bb)) {
	p->st.n_reject++;
	a_warn("PEER %s decrypt-failed", p->name);
	return;
      }
      if (BOK(&bb)) {
	p->st.n_ipin++;
	p->st.sz_ipin += BSZ(&b);
	tun_inject(&p->t, &bb);
      } else {
	p->st.n_reject++;
	a_warn("PEER %s packet-build-failed", p->name);
      }
      break;
    case MSG_KEYEXCH:
      kx_message(&p->kx, ch & MSG_TYPEMASK, &b);
      break;
    default:
      p->st.n_reject++;
      a_warn("PEER %s bad-packet unknown-category 0x%02x", p->name, ch);
      break;
  }
}

/* --- @p_txstart@ --- *
 *
 * Arguments:	@peer *p@ = pointer to peer block
 *		@unsigned msg@ = message type code
 *
 * Returns:	A pointer to a buffer to write to.
 *
 * Use:		Starts sending to a peer.  Only one send can happen at a
 *		time.
 */

buf *p_txstart(peer *p, unsigned msg)
{
  buf_init(&p->b, buf_o, sizeof(buf_o));
  buf_putbyte(&p->b, msg);
  return (&p->b);
}

/* --- @p_txend@ --- *
 *
 * Arguments:	@peer *p@ = pointer to peer block
 *
 * Returns:	---
 *
 * Use:		Sends a packet to the peer.
 */

void p_txend(peer *p)
{
  if (!BOK(&p->b)) {
    a_warn("PEER %s packet-build-failed", p->name);
    return;
  }
  IF_TRACING(T_PEER, trace_block(T_PACKET, "peer: sending packet",
				 BBASE(&p->b), BLEN(&p->b)); )
  if (sendto(sock.fd, BBASE(&p->b), BLEN(&p->b),
	     0, &p->peer.sa, p->sasz) < 0)
    a_warn("PEER %s socket-write-error -- %s", p->name, strerror(errno));
  else {
    p->st.n_out++;
    p->st.sz_out += BLEN(&p->b);
  }
}

/* --- @p_tun@ --- *
 *
 * Arguments:	@peer *p@ = pointer to peer block
 *		@buf *b@ = buffer containing incoming packet
 *
 * Returns:	---
 *
 * Use:		Handles a packet which needs to be sent to a peer.
 */

void p_tun(peer *p, buf *b)
{
  buf *bb = p_txstart(p, MSG_PACKET);

  TIMER;
  if (ksl_encrypt(&p->ks, MSG_PACKET, b, bb))
    kx_start(&p->kx);
  if (BOK(bb) && BLEN(bb)) {
    p->st.n_ipout++;
    p->st.sz_ipout += BLEN(bb);
    p_txend(p);
  }
}

/* --- @p_interval@ --- *
 *
 * Arguments:	---
 *
 * Returns:	---
 *
 * Use:		Called periodically to do tidying.
 */

void p_interval(void)
{
  peer *p, *pp;
  int reload;

  reload = km_interval();
  for (p = peers; p; p = pp) {
    pp = p->next;
    if (reload)
      kx_newkeys(&p->kx);
    ksl_prune(&p->ks);
  }
}

/* --- @p_stats@ --- *
 *
 * Arguments:	@peer *p@ = pointer to a peer block
 *
 * Returns:	A pointer to the peer's statistics.
 */

stats *p_stats(peer *p) { return (&p->st); }

/* --- @p_ifname@ --- *
 *
 * Arguments:	@peer *p@ = pointer to a peer block
 *
 * Returns:	A pointer to the peer's interface name.
 */

const char *p_ifname(peer *p) { return (tun_ifname(&p->t)); }

/* --- @p_addr@ --- *
 *
 * Arguments:	@peer *p@ = pointer to a peer block
 *
 * Returns:	A pointer to the peer's address.
 */

const addr *p_addr(peer *p) { return (&p->peer); }

/* --- @p_init@ --- *
 *
 * Arguments:	@struct in_addr addr@ = address to bind to
 *		@unsigned port@ = port number to listen to
 *
 * Returns:	---
 *
 * Use:		Initializes the peer system; creates the socket.
 */

void p_init(struct in_addr addr, unsigned port)
{
  int fd;
  struct sockaddr_in sin;
  int len = PKBUFSZ;

  /* --- Note on socket buffer sizes --- *
   *
   * For some bizarre reason, Linux 2.2 (at least) doubles the socket buffer
   * sizes I pass to @setsockopt@.  I'm not putting special-case code here
   * for Linux: BSD (at least TCPv2) does what I tell it rather than second-
   * guessing me.
   */

  if ((fd = socket(PF_INET, SOCK_DGRAM, 0)) < 0)
    die(EXIT_FAILURE, "socket creation failed: %s", strerror(errno));
  BURN(sin);
  sin.sin_family = AF_INET;
  sin.sin_addr = addr;
  sin.sin_port = htons(port);
  if (bind(fd, (struct sockaddr *)&sin, sizeof(sin)))
    die(EXIT_FAILURE, "bind failed: %s", strerror(errno));
  if (setsockopt(fd, SOL_SOCKET, SO_RCVBUF, &len, sizeof(len)) ||
      setsockopt(fd, SOL_SOCKET, SO_SNDBUF, &len, sizeof(len))) {
    die(EXIT_FAILURE, "failed to set socket buffer sizes: %s",
	strerror(errno));
  }
  fdflags(fd, O_NONBLOCK, O_NONBLOCK, FD_CLOEXEC, FD_CLOEXEC);
  sel_initfile(&sel, &sock, fd, SEL_READ, p_read, 0);
  sel_addfile(&sock);
  T( trace(T_PEER, "peer: created socket"); )
}

/* --- @p_port@ --- *
 *
 * Arguments:	---
 *
 * Returns:	Port number used for socket.
 */

unsigned p_port(void)
{
  addr a;
  size_t sz = sizeof(addr);

  if (getsockname(sock.fd, &a.sa, &sz))
    die(EXIT_FAILURE, "couldn't read port number: %s", strerror(errno));
  assert(a.sa.sa_family == AF_INET);
  return (ntohs(a.sin.sin_port));
}

/* --- @p_create@ --- *
 *
 * Arguments:	@const char *name@ = name for this peer
 *		@struct sockaddr *sa@ = socket address of peer
 *		@size_t sz@ = size of socket address
 *
 * Returns:	Pointer to the peer block, or null if it failed.
 *
 * Use:		Creates a new named peer block.  No peer is actually attached
 *		by this point.
 */

peer *p_create(const char *name, struct sockaddr *sa, size_t sz)
{
  peer *p = CREATE(peer);
  T( trace(T_PEER, "peer: creating new peer `%s'", name); )
  p->name = xstrdup(name);
  p->ks = 0;
  p->prev = 0;
  memcpy(&p->peer.sa, sa, sz);
  p->sasz = sz;
  memset(&p->st, 0, sizeof(stats));
  p->st.t_start = time(0);
  if (kx_init(&p->kx, p, &p->ks))
    goto tidy_0;
  if (tun_create(&p->t, p))
    goto tidy_1;
  p->next = peers;
  if (peers)
    peers->prev = p;
  peers = p;
  switch (p->peer.sa.sa_family) {
    case AF_INET:
      a_notify("ADD %s INET %s %u",
	       name,
	       inet_ntoa(p->peer.sin.sin_addr),
	       (unsigned)ntohs(p->peer.sin.sin_port));
      break;
    default:
      a_notify("ADD %s UNKNOWN", name);
      break;
  }
  return (p);

tidy_1:
  kx_free(&p->kx);
tidy_0:
  xfree(p->name);
  DESTROY(p);
  return (0);
}

/* --- @p_name@ --- *
 *
 * Arguments:	@peer *p@ = pointer to a peer block
 *
 * Returns:	A pointer to the peer's name.
 */

const char *p_name(peer *p) { return (p->name); }

/* --- @p_find@ --- *
 *
 * Arguments:	@const char *name@ = name to look up
 *
 * Returns:	Pointer to the peer block, or null if not found.
 *
 * Use:		Finds a peer by name.
 */

peer *p_find(const char *name)
{
  peer *p;
  for (p = peers; p; p = p->next) {
    if (strcmp(name, p->name) == 0)
      return (p);
  }
  return (0);  
}

/* --- @p_destroy@ --- *
 *
 * Arguments:	@peer *p@ = pointer to a peer
 *
 * Returns:	---
 *
 * Use:		Destroys a peer.
 */

void p_destroy(peer *p)
{
  T( trace(T_PEER, "peer: destroying peer `%s'", p->name); )
  a_notify("KILL %s", p->name);
  ksl_free(&p->ks);
  kx_free(&p->kx);
  tun_destroy(&p->t);
  xfree(p->name);
  if (p->next)
    p->next->prev = p->prev;
  if (p->prev)
    p->prev->next = p->next;
  else
    peers = p->next;
  DESTROY(p);
}

/* --- @p_first@, @p_next@ --- *
 *
 * Arguments:	@peer *p@ = a peer block
 *
 * Returns:	@peer_first@ returns the first peer in some ordering;
 *		@peer_next@ returns the peer following a given one in the
 *		same ordering.  Null is returned for the end of the list.
 */

peer *p_first(void) { return (peers); }
peer *p_next(peer *p) { return (p->next); }

/*----- That's all, folks -------------------------------------------------*/
